package data.scripts.themes;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.impl.campaign.procgen.StarSystemGenerator;
import com.fs.starfarer.api.impl.campaign.procgen.themes.ThemeGenerator;
import com.fs.starfarer.api.impl.campaign.procgen.themes.ThemeGenContext;
import com.fs.starfarer.api.impl.campaign.procgen.themes.MiscellaneousThemeGenerator;
import data.scripts.themes.DerelictThemeGeneratorMod;
import data.scripts.themes.RemnantThemeGeneratorMod;
import data.scripts.themes.RuinsThemeGeneratorMod;
import data.campaign.procgen.themes.BladeBreakerThemeGenerator;
import src.data.scripts.campaign.SAD_ThemeGenerator;

public class SectorThemeGenerator {

	public static List<ThemeGenerator> generators = new ArrayList<ThemeGenerator>();
	
	static {
		generators.add(new DerelictThemeGeneratorMod());
		generators.add(new RemnantThemeGeneratorMod());
		generators.add(new RuinsThemeGeneratorMod());
		generators.add(new MiscellaneousThemeGenerator());
		if (Global.getSettings().getModManager().isModEnabled("sad")) {
			generators.add(new SAD_ThemeGenerator());
		}
		if (Global.getSettings().getModManager().isModEnabled("istl_dam")) {
			generators.add(new BladeBreakerThemeGenerator());
		}
	}
	
	public static void generate(ThemeGenContext context) {
		Collections.sort(generators, new Comparator<ThemeGenerator>() {
			public int compare(ThemeGenerator o1, ThemeGenerator o2) {
				int result = o1.getOrder() - o2.getOrder();
				if (result == 0) return o1.getThemeId().compareTo(o2.getThemeId());
				return result;
			}
		});
		
		float totalWeight = 0f;
		for (ThemeGenerator g : generators) {
			totalWeight += g.getWeight();
			g.setRandom(StarSystemGenerator.random);
		}
		
		for (ThemeGenerator g : generators) {
			float w = g.getWeight();
			
			float f = 0f;
			if (totalWeight > 0) {
				f = w / totalWeight; 
			} else {
				if (w > 0) f = 1f;
			}
			//g.setRandom(StarSystemGenerator.random);
			g.generateForSector(context, f);
			
			//float used = context.majorThemes.size();
			totalWeight -= w;

		}
	}
	
	
	
	
}
